@@ -59,26 +59,26 @@
 #endif
 
 // --- LCD ---
-#ifndef WAVESHARE_RP2040_LCD_SPI
-#define WAVESHARE_RP2040_LCD_SPI 1
+#ifndef WAVESHARE_LCD_SPI
+#define WAVESHARE_LCD_SPI 1
 #endif
-#ifndef WAVESHARE_RP2040_LCD_DC_PIN
-#define WAVESHARE_RP2040_LCD_DC_PIN 8
+#ifndef WAVESHARE_LCD_DC_PIN
+#define WAVESHARE_LCD_DC_PIN 8
 #endif
-#ifndef WAVESHARE_RP2040_LCD_CS_PIN
-#define WAVESHARE_RP2040_LCD_CS_PIN 9
+#ifndef WAVESHARE_LCD_CS_PIN
+#define WAVESHARE_LCD_CS_PIN 9
 #endif
-#ifndef WAVESHARE_RP2040_LCD_SCLK_PIN
-#define WAVESHARE_RP2040_LCD_SCLK_PIN 10
+#ifndef WAVESHARE_LCD_SCLK_PIN
+#define WAVESHARE_LCD_SCLK_PIN 10
 #endif
-#ifndef WAVESHARE_RP2040_LCD_TX_PIN
-#define WAVESHARE_RP2040_LCD_TX_PIN 11
+#ifndef WAVESHARE_LCD_TX_PIN
+#define WAVESHARE_LCD_TX_PIN 11
 #endif
-#ifndef WAVESHARE_RP2040_LCD_RST_PIN
-#define WAVESHARE_RP2040_LCD_RST_PIN 12
+#ifndef WAVESHARE_LCD_RST_PIN
+#define WAVESHARE_LCD_RST_PIN 12
 #endif
-#ifndef WAVESHARE_RP2040_LCD_BL_PIN
-#define WAVESHARE_RP2040_LCD_BL_PIN 25
+#ifndef WAVESHARE_LCD_BL_PIN
+#define WAVESHARE_LCD_BL_PIN 25
 #endif
 
 // --- FLASH ---
